# -*- coding: utf-8 -*-
"""
batterydataextractor.doc.text

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Text-based document elements.
"""
from abc import abstractmethod, ABC
import collections
import logging
import re
import unicodedata
import six

from ..model.base import ModelList
from ..model.model import Compound
from ..nlp.lexicon import ChemLexicon, Lexicon
from ..nlp.cem import CemTagger
from ..nlp.abbrev import AbbreviationDetector
from ..nlp.tag import NoneTagger, BaseTagger, BertTagger
from ..nlp.tokenize import ChemSentenceTokenizer, ChemWordTokenizer, SentenceTokenizer, WordTokenizer
from ..utils import memoized_property
from .element import BaseElement
from ..text import CONTROL_RE


log = logging.getLogger(__name__)


class BaseText(BaseElement):
    """Abstract base class for a text Document Element."""

    def __init__(self, text, word_tokenizer=None, lexicon=None, abbreviation_detector=None, pos_tagger=None, ner_tagger=None, **kwargs):
        """
        .. note::
            If intended as part of a :class:`~batterydataextractor.doc.document.Document`,
            an element should either be initialized with a reference to its containing document,
            or its :attr:`document` attribute should be set as soon as possible.
            If the element is being passed in to a :class:`~batterydataextractor.doc.document.Document`
            to initialise it, the :attr:`document` attribute is automatically set
            during the initialisation of the document, so the user does not need to worry about this.
        :param str text: The text contained in this element.
        :param WordTokenizer word_tokenizer: (Optional) Word tokenizer for this element.
        :param Lexicon lexicon: (Optional) Lexicon for this element. The lexicon stores all the occurences of unique words and can provide
            Brown clusters for the words.
        :param AbbreviationDetector abbreviation_detector: (Optional) The abbreviation detector for this element.
        :param BaseTagger pos_tagger: (Optional) The part of speech tagger for this element.
        :param BaseTagger ner_tagger: (Optional) The named entity recognition tagger for this element.
        :param Document document: (Optional) The document containing this element.
        :param str label: (Optional) The label for the captioned element, e.g. Table 1 would have a label of 1.
        :param Any id: (Optional) Some identifier for this element. Must be equatable.
        :param list[batterydataextractor.models.BaseModel] models: (Optional) A list of models for this element to parse.
            If the element is part of another element (e.g. a :class:`~batterydataextractor.doc.text.Sentence`
            inside a :class:`~batterydataextractor.doc.text.Paragraph`), or is part of a :class:`~batterydataextractor.doc.document.Document`,
            this is set automatically to be the same as that of the containing element, unless manually set otherwise.
        """
        if not isinstance(text, six.text_type):
            raise TypeError('Text must be a unicode string')
        super(BaseText, self).__init__(**kwargs)
        self._text = unicodedata.normalize("NFKD", text)
        self.word_tokenizer = word_tokenizer if word_tokenizer is not None else self.word_tokenizer
        self.lexicon = lexicon if lexicon is not None else self.lexicon
        self.abbreviation_detector = abbreviation_detector if abbreviation_detector is not None else self.abbreviation_detector
        self.pos_tagger = pos_tagger if pos_tagger is not None else self.pos_tagger
        self.ner_tagger = ner_tagger if ner_tagger is not None else self.ner_tagger

    def __repr__(self):
        return '%s(id=%r, references=%r, text=%r)' % (self.__class__.__name__, self.id, self.references, self._text)

    def __str__(self):
        return self._text

    @property
    def text(self):
        """The raw text :class:`str` for this passage of text."""
        return self._text

    @property
    @abstractmethod
    def word_tokenizer(self):
        """The :class:`~batterydataextractor.nlp.tokenize.WordTokenizer` used by this element."""
        return

    @property
    @abstractmethod
    def lexicon(self):
        """The :class:`~batterydataextractor.nlp.lexicon.Lexicon` used by this element."""
        return

    @property
    @abstractmethod
    def pos_tagger(self):
        """The part of speech tagger used by this element. A subclass of :class:`~batterydataextractor.nlp.tag.BaseTagger`"""
        return

    @property
    @abstractmethod
    def ner_tagger(self):
        """The named entity recognition tagger used by this element. A subclass of :class:`~batterydataextractor.nlp.tag.BaseTagger`"""
        return

    @property
    @abstractmethod
    def tokens(self):
        """A list of :class:`Token` s for this object."""
        return

    @property
    @abstractmethod
    def tags(self):
        """
        A list of tags corresponding to each of the tokens in the object.
        For information on what each of the tags can be, check the documentation on
        the specific :attr:`ner_tagger` and :attr:`pos_tagger` used for this class.
        """
        return

    def serialize(self):
        """
        Convert self to a dictionary. The key 'type' will contain
        the name of the class being serialized, and the key 'content' will contain
        a serialized representation of :attr:`text`, which is a :class:`str`
        """
        data = {'type': self.__class__.__name__, 'content': self.text}
        return data

    def _repr_html_(self):
        return self.text

    @word_tokenizer.setter
    def word_tokenizer(self, value):
        self._word_tokenizer = value

    @lexicon.setter
    def lexicon(self, value):
        self._lexicon = value

    @pos_tagger.setter
    def pos_tagger(self, value):
        self._pos_tagger = value

    @ner_tagger.setter
    def ner_tagger(self, value):
        self._ner_tagger = value


class Text(collections.Sequence, BaseText):
    """A passage of text, comprising one or more sentences."""

    sentence_tokenizer = ChemSentenceTokenizer()
    word_tokenizer = ChemWordTokenizer()
    lexicon = ChemLexicon()
    abbreviation_detector = AbbreviationDetector()
    pos_tagger = BertTagger()
    ner_tagger = CemTagger()

    def __init__(self, text, sentence_tokenizer=None, word_tokenizer=None, lexicon=None, abbreviation_detector=None, pos_tagger=None, ner_tagger=None, parsers=None, **kwargs):
        """
        .. note::
            If intended as part of a :class:`~batterydataextractor.doc.document.Document`,
            an element should either be initialized with a reference to its containing document,
            or its :attr:`document` attribute should be set as soon as possible.
            If the element is being passed in to a :class:`~batterydataextractor.doc.document.Document`
            to initialise it, the :attr:`document` attribute is automatically set
            during the initialisation of the document, so the user does not need to worry about this.
        :param str text: The text contained in this element.
        :param SentenceTokenizer sentence_tokenizer: (Optional) Sentence tokenizer for this element.
            Default :class:`~batterydataextractor.nlp.tokenize.ChemSentenceTokenizer`.
        :param WordTokenizer word_tokenizer: (Optional) Word tokenizer for this element.
            Default :class:`~batterydataextractor.nlp.tokenize.ChemWordTokenizer`.
        :param Lexicon lexicon: (Optional) Lexicon for this element. The lexicon stores all the occurences of unique words and can provide
            Brown clusters for the words. Default :class:`~batterydataextractor.nlp.lexicon.ChemLexicon`
        :param AbbreviationDetector abbreviation_detector: (Optional) The abbreviation detector for this element.
            Default :class:`~batterydataextractor.nlp.abbrev.ChemAbbreviationDetector`.
        :param BaseTagger pos_tagger: (Optional) The part of speech tagger for this element.
            Default :class:`~batterydataextractor.nlp.pos.ChemCrfPosTagger`.
        :param BaseTagger ner_tagger: (Optional) The named entity recognition tagger for this element.
            Default :class:`~batterydataextractor.nlp.cem.CemTagger`
        :param Document document: (Optional) The document containing this element.
        :param str label: (Optional) The label for the captioned element, e.g. Table 1 would have a label of 1.
        :param Any id: (Optional) Some identifier for this element. Must be equatable.
        :param list[batterydataextractor.models.BaseModel] models: (Optional) A list of models for this element to parse.
            If the element is part of another element (e.g. a :class:`~batterydataextractor.doc.text.Sentence`
            inside a :class:`~batterydataextractor.doc.text.Paragraph`), or is part of a :class:`~batterydataextractor.doc.document.Document`,
            this is set automatically to be the same as that of the containing element, unless manually set otherwise.
        """
        super(Text, self).__init__(text, word_tokenizer=word_tokenizer, lexicon=lexicon,
                                   abbreviation_detector=abbreviation_detector, pos_tagger=pos_tagger,
                                   ner_tagger=ner_tagger, parsers=None, **kwargs)
        self.sentence_tokenizer = sentence_tokenizer if sentence_tokenizer is not None else self.sentence_tokenizer
        self.pos_tagger.device = self.device
        self.ner_tagger.device = self.device

    def __getitem__(self, index):
        return self.sentences[index]

    def __len__(self):
        return len(self.sentences)

    def set_config(self):
        """ Load settings from configuration file
        .. note:: Called when Document instance is created
        """

        if self.document is None:
            pass
        else:
            c = self.document.config
            if 'SENTENCE_TOKENIZER' in c.keys():
                self.sentence_tokenizer = eval(c['SENTENCE_TOKENIZER'])()
            if 'WORD_TOKENIZER' in c.keys():
                self.word_tokenizer = eval(c['WORD_TOKENIZER'])()
            if 'POS_TAGGER' in c.keys():
                self.pos_tagger = eval(c['POS_TAGGER'])()
            if 'NER_TAGGER' in c.keys():
                self.ner_tagger = eval(c['NER_TAGGER'])()
            if 'LEXICON' in c.keys():
                self.lexicon = eval(c['LEXICON'])()
            if 'PARSERS' in c.keys():
                raise(DeprecationWarning('Manually setting parsers deprecated, any settings from config files for this will be ignored.'))

    @memoized_property
    def sentences(self):
        """A list of :class:`Sentence` s that make up this text passage."""
        sents = []
        spans = self.sentence_tokenizer.span_tokenize(self.text)
        for span in spans:
            sent = Sentence(
                text=self.text[span[0]:span[1]],
                start=span[0],
                end=span[1],
                word_tokenizer=self.word_tokenizer,
                lexicon=self.lexicon,
                abbreviation_detector=self.abbreviation_detector,
                pos_tagger=self.pos_tagger,
                ner_tagger=self.ner_tagger,
                document=self.document,
                models=self.models,
                device=self.device
            )
            sents.append(sent)
        return sents

    @property
    def raw_sentences(self):
        """A list of :class:`str` for the sentences that make up this text passage."""
        return [sentence.text for sentence in self.sentences]

    @property
    def tokens(self):
        return [sent.tokens for sent in self.sentences]

    @property
    def raw_tokens(self):
        """A list of :class:`str` representations for the tokens of each sentence in this text passage."""
        return [sent.raw_tokens for sent in self.sentences]

    @property
    def pos_tagged_tokens(self):
        """A list of (:class:`Token` token, :class:`str` tag) tuples for each sentence in this text passage."""
        return [sent.pos_tagged_tokens for sent in self.sentences]

    @property
    def pos_tags(self):
        """A list of :class:`str` part of speech tags for each sentence in this text passage."""
        return [sent.pos_tags for sent in self.sentences]

    @memoized_property
    def unprocessed_ner_tagged_tokens(self):
        """
        A list of (:class:`Token` token, :class:`str` named entity recognition tag)
        from the text.
        No corrections from abbreviation detection are performed.
        """
        return [sent.unprocessed_ner_tagged_tokens for sent in self.sentences]

    @memoized_property
    def unprocessed_ner_tags(self):
        """
        A list of :class:`str` unprocessed named entity tags for the tokens in this sentence.
        No corrections from abbreviation detection are performed.
        """
        return [sent.unprocessed_ner_tags for sent in self.sentences]

    @property
    def ner_tagged_tokens(self):
        """
        A list of (:class:`Token` token, :class:`str` named entity recognition tag)
        from the text.
        """
        return [sent.ner_tagged_tokens for sent in self.sentences]

    @property
    def ner_tags(self):
        """
        A list of named entity tags corresponding to each of the tokens in the object.
        For information on what each of the tags can be, check the documentation on
        the specific :attr:`ner_tagger` used for this object.
        """
        return [sent.ner_tags for sent in self.sentences]

    @property
    def cems(self):
        """
        A list of all Chemical Entity Mentions in this text as :class:`batterydataextractor.doc.text.span`
        """
        return [cem for sent in self.sentences for cem in sent.cems]

    @property
    def tagged_tokens(self):
        """
        A list of (:class:`Token` token, :class:`str` named entity recognition tag)
        from the text.
        """
        return [sent.tagged_tokens for sent in self.sentences]

    @property
    def tags(self):
        return [sent.tags for sent in self.sentences]

    @property
    def abbreviation_definitions(self):
        """
        A list of all abbreviation definitions in this Document. Each abbreviation is in the form
        (:class:`str` abbreviation, :class:`str` long form of abbreviation, :class:`str` ner_tag)
        """
        return [ab for sent in self.sentences for ab in sent.abbreviation_definitions if ab != []]

    @property
    def records(self):
        """All records found in the object, as a list of :class:`~batterydataextractor.model.base.BaseModel`."""
        return ModelList(*[r for sent in self.sentences for r in sent.records])

    def __add__(self, other):
        if type(self) == type(other):
            merged = self.__class__(
                text=self.text + other.text,
                id=self.id or other.id,
                references=self.references + other.references,
                sentence_tokenizer=self.sentence_tokenizer,
                word_tokenizer=self.word_tokenizer,
                lexicon=self.lexicon,
                abbreviation_detector=self.abbreviation_detector,
                pos_tagger=self.pos_tagger,
                ner_tagger=self.ner_tagger,
            )
            return merged
        return NotImplemented


class Title(Text, ABC):

    def __init__(self, text, **kwargs):
        super(Title, self).__init__(text, **kwargs)
        self.models = [Compound]

    def _repr_html_(self):
        return '<h1 class="bde-title">' + self.text + '</h1>'


class Heading(Text):

    def __init__(self, text, **kwargs):
        super(Heading, self).__init__(text, **kwargs)
        self.models = [Compound]

    def _repr_html_(self):
        return '<h2 class="bde-title">' + self.text + '</h2>'


class Paragraph(Text):

    def __init__(self, text, **kwargs):
        super(Paragraph, self).__init__(text, **kwargs)
        self.models = [Compound]

    def _repr_html_(self):
        return '<p class="bde-paragraph">' + self.text + '</p>'


class Footnote(Text):

    def __init__(self, text, **kwargs):
        super(Footnote, self).__init__(text, **kwargs)

    def _repr_html_(self):
        return '<p class="bde-footnote">' + self.text + '</p>'


class Citation(Text):
    ner_tagger = NoneTagger()  #: No tagging is done for citations
    abbreviation_detector = None

    def _repr_html_(self):
        return '<p class="bde-citation">' + self.text + '</p>'


class Caption(Text):

    def __init__(self, text, **kwargs):
        super(Caption, self).__init__(text, **kwargs)
        self.models = [Compound]

    def _repr_html_(self):
        return '<caption class="bde-caption">' + self.text + '</caption>'


class Abstract(Text):
    def __init__(self, text, **kwargs):
        super(Abstract, self).__init__(text, **kwargs)
        self.models = [Compound]

    def _repr_html_(self):
        return '<h2 class="bde-abstract">' + self.text + '</h2>'


class Sentence(BaseText, ABC):
    """A single sentence within a text passage."""

    word_tokenizer = ChemWordTokenizer()
    lexicon = ChemLexicon()
    abbreviation_detector = AbbreviationDetector()
    pos_tagger = BertTagger()
    ner_tagger = CemTagger()

    def __init__(self, text, start=0, end=None, word_tokenizer=None, lexicon=None, abbreviation_detector=None, pos_tagger=None, ner_tagger=None, **kwargs):
        """
        .. note::
            If intended as part of a :class:`batterydataextractor.doc.document.Document`,
            an element should either be initialized with a reference to its containing document,
            or its :attr:`document` attribute should be set as soon as possible.
            If the element is being passed in to a :class:`batterydataextractor.doc.document.Document`
            to initialise it, the :attr:`document` attribute is automatically set
            during the initialisation of the document, so the user does not need to worry about this.
        :param str text: The text contained in this element.
        :param int start: (Optional) The starting index of the sentence within the containing element. Default 0.
        :param int end: (Optional) The end index of the sentence within the containing element. Defualt None
        :param WordTokenizer word_tokenizer: (Optional) Word tokenizer for this element.
            Default :class:`~batterydataextractor.nlp.tokenize.ChemWordTokenizer`.
        :param Lexicon lexicon: (Optional) Lexicon for this element. The lexicon stores all the occurences of unique words and can provide
            Brown clusters for the words. Default :class:`~batterydataextractor.nlp.lexicon.ChemLexicon`
        :param AbbreviationDetector abbreviation_detector: (Optional) The abbreviation detector for this element.
            Default :class:`~batterydataextractor.nlp.abbrev.AbbreviationDetector`.
        :param BaseTagger pos_tagger: (Optional) The part of speech tagger for this element.
            Default :class:`~batterydataextractor.nlp.pos.ChemCrfPosTagger`.
        :param BaseTagger ner_tagger: (Optional) The named entity recognition tagger for this element.
            Default :class:`~batterydataextractor.nlp.cem.CemTagger`
        :param Document document: (Optional) The document containing this element.
        :param str label: (Optional) The label for the captioned element, e.g. Table 1 would have a label of 1.
        :param Any id: (Optional) Some identifier for this element. Must be equatable.
        :param list[batterydataextractor.models.BaseModel] models: (Optional) A list of models for this element to parse.
            If the element is part of another element (e.g. a :class:`~batterydataextractor.doc.text.Sentence`
            inside a :class:`~batterydataextractor.doc.text.Paragraph`), or is part of a :class:`~batterydataextractor.doc.document.Document`,
            this is set automatically to be the same as that of the containing element, unless manually set otherwise.
        """
        self.models = [Compound]
        super(Sentence, self).__init__(text, word_tokenizer=word_tokenizer, lexicon=lexicon, abbreviation_detector=abbreviation_detector, pos_tagger=pos_tagger, ner_tagger=ner_tagger, **kwargs)
        #: The start index of this sentence within the text passage.
        self.start = start
        #: The end index of this sentence within the text passage.
        self.end = end if end is not None else len(text)

    def __repr__(self):
        return '%s(%r, %r, %r)' % (self.__class__.__name__, self._text, self.start, self.end)

    @memoized_property
    def tokens(self):
        spans = self.word_tokenizer.span_tokenize(self.text)
        toks = [Token(
            text=self.text[span[0]:span[1]],
            start=span[0] + self.start,
            end=span[1] + self.start,
            lexicon=self.lexicon
        ) for span in spans]
        return toks

    @property
    def raw_tokens(self):
        """A list of :class:`str` representations for the tokens in the object."""
        return [token.text for token in self.tokens]

    @memoized_property
    def pos_tagged_tokens(self):
        """A list of (:class:`Token` token, :class:`str` tag) tuples for each sentence in this sentence."""
        # log.debug('Getting pos tags')
        return self.pos_tagger.tag(self.raw_tokens)

    @property
    def pos_tags(self):
        """A list of :class:`str` part of speech tags for each sentence in this sentence."""
        return [tag for token, tag in self.pos_tagged_tokens]

    @memoized_property
    def unprocessed_ner_tagged_tokens(self):
        """
        A list of (:class:`Token` token, :class:`str` named entity recognition tag)
        from the text.
        No corrections from abbreviation detection are performed.
        """
        return self.ner_tagger.tag(self.pos_tagged_tokens)

    @memoized_property
    def unprocessed_ner_tags(self):
        """
        A list of :class:`str` unprocessed named entity tags for the tokens in this sentence.
        No corrections from abbreviation detection are performed.
        """
        return [tag for token, tag in self.unprocessed_ner_tagged_tokens]

    @memoized_property
    def abbreviation_definitions(self):
        """
        A list of all abbreviation definitions in this Document. Each abbreviation is in the form
        (:class:`str` abbreviation, :class:`str` long form of abbreviation, :class:`str` ner_tag)
        """
        abbreviations, long_words = [], []
        if self.abbreviation_detector:
            log.debug('Detecting abbreviations')
            abbr_spans, long_spans = self.abbreviation_detector.detect_spans(self.raw_tokens)
            if abbr_spans:
                for abbr_span in abbr_spans:
                    abb = " ".join(self.raw_tokens)[abbr_span[0]:abbr_span[1]]
                    abbr = ("Abbr: ", abb)
                    abbreviations.append(abbr)
            if long_spans:
                for long_span in long_spans:
                    lon = " ".join(self.raw_tokens)[long_span[0]:long_span[1]]
                    long = ("LF: ", lon)
                    long_words.append(long)
        return abbreviations, long_words

    @memoized_property
    def ner_tagged_tokens(self):
        """
        A list of (:class:`Token` token, :class:`str` named entity recognition tag)
        from the sentence.
        """
        return list(zip(self.raw_tokens, self.ner_tags))

    @memoized_property
    def ner_tags(self):
        """
        A list of named entity tags corresponding to each of the tokens in the object.
        For information on what each of the tags can be, check the documentation on
        the specific :attr:`ner_tagger` used for this object.
        """
        # log.debug('Getting ner_tags')
        ner_tags = self.unprocessed_ner_tags
        return ner_tags

    @memoized_property
    def cems(self):
        """
        A list of all Chemical Entity Mentions in this text as :class:`~batterydataextractor.doc.text.Span`
        """
        log.debug('Getting cems')
        spans = []
        raw_tokens = self.raw_tokens
        for index, result in enumerate(self.ner_tags):
            if result == 'MAT':
                ner_word = raw_tokens[index].replace("(", "\\(").replace(")", "\\)")
                span = Span(text=raw_tokens[index], start=re.search(ner_word, self.text).start() + self.start,
                            end=re.search(ner_word, self.text).end() + self.start)
                spans.append(span)
        return spans

    @memoized_property
    def tags(self):
        tags = self.pos_tags
        for i, tag in enumerate(self.ner_tags):
            if tag is not None:
                tags[i] = tag
        return tags

    @property
    def tagged_tokens(self):
        """
        A list of (:class:`Token` token, :class:`str` named entity recognition tag)
        from the text.
        """
        return list(zip(self.raw_tokens, self.tags))

    @property
    def records(self):
        """All records found in the object, as a list of :class:`~batterydataextractor.model.base.BaseModel`."""
        records = ModelList()
        # Ensure no control characters are sent to a parser (need to be XML compatible)
        tagged_tokens = [(CONTROL_RE.sub('', token), tag) for token, tag in self.tagged_tokens]
        for model in self._streamlined_models:
            for parser in model.parsers:
                if hasattr(parser, 'parse_sentence'):
                    for record in parser.parse_sentence(tagged_tokens):
                        p = record.serialize()
                        if not p:  # TODO: Potential performance issues?
                            continue
                        # Skip duplicate records
                        if record in records:
                            continue
                        records.append(record)
        # i = 0
        # length = len(records)
        # while i < length:
        #     j = 0
        #     while j < length:
        #         if i != j:
        #             records[j].merge_all(records[i])
        #         j += 1
        #     i += 1
        return records

    def __add__(self, other):
        if type(self) == type(other):
            merged = self.__class__(
                text=self.text + other.text,
                start=self.start,
                end=None,
                id=self.id or other.id,
                references=self.references + other.references,
                word_tokenizer=self.word_tokenizer,
                lexicon=self.lexicon,
                abbreviation_detector=self.abbreviation_detector,
                pos_tagger=self.pos_tagger,
                ner_tagger=self.ner_tagger,
            )
            return merged
        return NotImplemented


class Span(object):
    """A text span within a sentence."""

    def __init__(self, text, start, end):
        """
        :param str text: The text contained by this span.
        :param int start: The start offset of this token in the original text.
        :param int end: The end offsent of this token in the original text.
        """
        self.text = text
        """The :class:`str` text content of this span."""
        self.start = start
        """The :class:`int` start offset of this token in the original text."""
        self.end = end
        """The :class:`int` end offset of this token in the original text."""

    def __repr__(self):
        return '%s(%r, %r, %r)' % (self.__class__.__name__, self.text, self.start, self.end)

    def __str__(self):
        return self.text

    def __eq__(self, other):
        """Span objects are equal if the source text is equal, and the start and end indices are equal."""
        if not isinstance(other, self.__class__):
            return False
        return self.text == other.text and self.start == other.start and self.end == other.end

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash((self.text, self.start, self.end))

    @property
    def length(self):
        """The :class:`int` offset length of this span in the original text."""
        return self.end - self.start


class Token(Span):
    """A single token within a sentence. Corresponds to a word, character, punctuation etc."""

    def __init__(self, text, start, end, lexicon):
        """
        :param str text: The text contained by this token.
        :param int start: The start offset of this token in the original text.
        :param int end: The end offsent of this token in the original text.
        :param Lexicon lexicon: The lexicon which contains this token.
        """
        super(Token, self).__init__(text, start, end)
        #: The lexicon for this token.
        self.lexicon = lexicon
        self.lexicon.add(text)

    @property
    def lex(self):
        """The corresponding :class:`batterydataextractor.nlp.lexicon.Lexeme` entry in the Lexicon for this token."""
        return self.lexicon[self.text]



from .tag import BertTagger, BaseTagger
from .cem import BertCemTagger, CemTagger
from .abbrev import AbbreviationDetector
from .lexicon import ChemLexicon
from .tokenize import SentenceTokenizer, ChemSentenceTokenizer, ChemWordTokenizer, FineWordTokenizer, WordTokenizer



import logging
import unittest

from batterydataextractor.doc.document import Document

logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)


class TestDocument(unittest.TestCase):
    """Simple Document instantiation tests."""

    def test_text_elements(self):
        """Test Document can be instantiated with a list of strings."""
        els = [
            'A first paragraph. With two sentences.',
            'A second paragraph.',
            'A third paragraph.'
        ]
        d = Document(*els)
        self.assertEqual(d.elements[0].text, 'A first paragraph. With two sentences.')
        self.assertEqual(d.elements[0].sentences[1].text, 'With two sentences.')
        self.assertEqual(d.elements[1].document, d)

    def test_bytestring_elements(self):
        """Test Document can be instantiated with a list of bytestrings."""
        els = [
            'A first paragraph. With two sentences.'.encode('ascii'),
            'A second paragraph. \u00a9'.encode('utf-8'),
            'A third paragraph (\u00b6).'.encode('windows-1252'),
        ]
        d = Document(*els)
        self.assertEqual(d.elements[0].text, 'A first paragraph. With two sentences.')
        self.assertEqual(d.elements[0].sentences[1].text, 'With two sentences.')
        self.assertEqual(d.elements[1].document, d)

    def test_document_iter(self):
        """Test Document can be iterated like a list to access its elements."""
        els = [
            'A first paragraph. With two sentences.',
            'A second paragraph.',
            'A third paragraph.'
        ]
        d = Document(*els)
        self.assertEqual(len(d), 3)
        self.assertEqual(d[2].text, 'A third paragraph.')
        self.assertEqual([e.text for e in d], els)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.errors

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Error classes for BatteryDataExtractor.
"""


class BatteryDataExtractorError(Exception):
    """Base ChemDataExtractor exception."""
    pass


class ReaderError(BatteryDataExtractorError):
    """Raised when a reader is unable to read a document."""


class ModelNotFoundError(BatteryDataExtractorError):
    """Raised when a model file could not be found."""



"""
batterydataextractor.__init__

BatteryDataExtractor
"""

__version__ = "1.0.0"
__author__ = 'Shu Huang'
__credits__ = 'Molecular Engineering Group, Cavendish Laboratory'

from .doc.document import Document



import unittest
from batterydataextractor.nlp.tokenize import WordTokenizer, FineWordTokenizer, ChemWordTokenizer
from batterydataextractor.doc import Text


class TestWordTokenizer(unittest.TestCase):
    """Test the standard word tokenizer."""

    maxDiff = None

    def setUp(self):
        self.t = WordTokenizer()

    def test_final_full_stop(self):
        """Test the word tokenizer splits off final full stop only."""
        self.assertEqual(
            ['This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.'],
            self.t.tokenize('This is Mr. Hoppy\'s tortoise.')
        )

    def test_full_stop_following(self):
        """Test the word tokenizer splits off final full stop if followed by brackets or quotes."""
        self.assertEqual(
            ['(', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.', ')'],
            self.t.tokenize('(This is Mr. Hoppy\'s tortoise.)')
        )
        self.assertEqual(
            ['"', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.', '"'],
            self.t.tokenize('"This is Mr. Hoppy\'s tortoise."')
        )
        self.assertEqual(
            ['"', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tort.oise', '.', '"'],
            self.t.tokenize('"This is Mr. Hoppy\'s tort.oise."')
        )

    def test_dollar(self):
        """Test the word tokenizer on dollar symbol."""
        self.assertEqual(
            ['On', 'a', '$', '50,000', 'mortgage', 'of', '30', 'years', 'at', '8', 'percent', ',', 'the', 'monthly', 'payment', 'would', 'be', '$', '366.88', '.'],
            self.t.tokenize('On a $50,000 mortgage of 30 years at 8 percent, the monthly payment would be $366.88.')
        )

    def test_quote(self):
        """Test the word tokenizer on quotes."""
        self.assertEqual(
            ['"', 'We', 'beat', 'some', 'pretty', 'good', 'teams', 'to', 'get', 'here', ',', '"', 'Slocum', 'said', '.'],
            self.t.tokenize('"We beat some pretty good teams to get here," Slocum said.')
        )

    def test_brackets_quotes(self):
        """Test the word tokenizer on brackets and quotes."""
        self.assertEqual(
            ['Well', ',', 'we', 'could', 'n\'t', 'have', 'this', 'predictable', ',', 'cliche', '-', 'ridden', ',', '"', 'Touched', 'by', 'an', 'Angel', '"', '(', 'a', 'show', 'creator', 'John', 'Masius', 'worked', 'on', ')', 'wanna-be', 'if', 'she', 'did', 'n\'t', '.'],
            self.t.tokenize('Well, we couldn\'t have this predictable, cliche-ridden, "Touched by an Angel" (a show creator John Masius worked on) wanna-be if she didn\'t.')
        )

    def test_exclamation(self):
        """Test the word tokenizer on exclamation mark."""
        self.assertEqual(
            ['I', 'can', 'not', 'can', 'not', 'work', 'under', 'these', 'conditions', '!'],
            self.t.tokenize('I cannot cannot work under these conditions!')
        )

    def test_digit_comma(self):
        """Test the word tokenizer on commas within numbers."""
        self.assertEqual(
            ['The', 'company', 'spent', '$', '30,000,000', 'last', 'year', '.'],
            self.t.tokenize('The company spent $30,000,000 last year.')
        )

    def test_decimal_number(self):
        """Test the word tokenizer on number containing full stop."""
        self.assertEqual(
            ['It', '\'s',  '2.45',  'cats', 'per', 'mango', '.'],
            self.t.tokenize('It\'s 2.45 cats per mango.')
        )

    def test_phone_number(self):
        """Test the word tokenizer on phone number containing hyphens"""
        self.assertEqual(
            ['Call',  'me',  'at', '02-2348-2192', '.'],
            self.t.tokenize('Call me at 02-2348-2192.')
        )

    def test_percentage(self):
        """Test the word tokenizer on percent sign."""
        self.assertEqual(
            ['The', 'company', 'spent', '40.75', '%', 'of', 'its', 'income', 'last', 'year', '.'],
            self.t.tokenize('The company spent 40.75% of its income last year.')
        )

    def test_colon_time(self):
        """Test the word tokenizer on colon between digits in a time."""
        self.assertEqual(
            ['He', 'arrived', 'at', '3:00', 'pm', '.'],
            self.t.tokenize('He arrived at 3:00 pm.')
        )

    def test_word_colon(self):
        """Test the word tokenizer on colon after word."""
        self.assertEqual(
            ['I', 'bought', 'these', 'items', ':', 'books', ',', 'pencils', ',', 'and', 'pens', '.'],
            self.t.tokenize('I bought these items: books, pencils, and pens.')
        )

    def test_digit_comma_space(self):
        """Test the word tokenizer on comma between digits with a space."""
        self.assertEqual(
            ['Though', 'there', 'were', '150', ',', '100', 'of', 'them', 'were', 'old', '.'],
            self.t.tokenize('Though there were 150, 100 of them were old.')
        )

    def test_digit_comma_multiple(self):
        """Test the word tokenizer on comma at end of digits."""
        self.assertEqual(
            ['There', 'were', '300,000', ',', 'but', 'that', 'was', "n't", 'enough', '.'],
            self.t.tokenize('There were 300,000, but that wasn\'t enough.')
        )

    def test_theyll(self):
        """Test the word tokenizer on the word they'll."""
        self.assertEqual(
            ['They', "'ll", 'save', 'and', 'invest', 'more', '.'],
            self.t.tokenize('They\'ll save and invest more.')
        )

    def test_bracket1(self):
        """Test the word tokenizer on sentence containing brackets."""
        self.assertEqual(
            ['For', 'a', 'few', 'weeks', '(', '>', '24', 'days', ')', '.'],
            self.t.tokenize('For a few weeks (>24 days).')
        )

    def test_contractions(self):
        """Test the word tokenizer on contractions."""
        self.assertEqual(
            ['Lem', 'me', 'in', ',', 'I', 'got', 'ta', 'gim', 'me', 'some', 'things', '.'],
            self.t.tokenize('Lemme in, I gotta gimme some things.')
        )

    def test_url(self):
        """Test the word tokenizer on URL."""
        self.assertEqual(
            ['The', 'address', 'is', 'http://www.chemdataextractor.org'],
            self.t.tokenize('The address is http://www.chemdataextractor.org')
        )

    def test_text_sentence(self):
        """Test tokenization through the Text and Sentence API."""
        t = Text('Hi, my name is Matt. What is your name?', word_tokenizer=WordTokenizer())
        self.assertEqual(
            [['Hi', ',', 'my', 'name', 'is', 'Matt', '.'], ['What', 'is', 'your', 'name', '?']],
            [sent.raw_tokens for sent in t.sentences]
        )


class TestChemTokenizer(unittest.TestCase):
    """Test the chemistry-aware word tokenizer."""

    maxDiff = None

    def setUp(self):
        self.t = ChemWordTokenizer()

    def test_final_full_stop(self):
        """Test the word tokenizer splits off final full stop only."""
        self.assertEqual(
            ['This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.'],
            self.t.tokenize('This is Mr. Hoppy\'s tortoise.')
        )

    def test_full_stop_following(self):
        """Test the word tokenizer splits off final full stop if followed by brackets or quotes."""
        self.assertEqual(
            ['(', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.', ')'],
            self.t.tokenize('(This is Mr. Hoppy\'s tortoise.)')
        )
        self.assertEqual(
            ['"', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.', '"'],
            self.t.tokenize('"This is Mr. Hoppy\'s tortoise."')
        )
        self.assertEqual(
            ['"', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tort.oise', '.', '"'],
            self.t.tokenize('"This is Mr. Hoppy\'s tort.oise."')
        )

    def test_dollar(self):
        """Test the word tokenizer on dollar symbol."""
        self.assertEqual(
            ['On', 'a', '$', '50,000', 'mortgage', 'of', '30', 'years', 'at', '8', 'percent', ',', 'the', 'monthly', 'payment', 'would', 'be', '$', '366.88', '.'],
            self.t.tokenize('On a $50,000 mortgage of 30 years at 8 percent, the monthly payment would be $366.88.')
        )

    def test_quote(self):
        """Test the word tokenizer on quotes."""
        self.assertEqual(
            ['"', 'We', 'beat', 'some', 'pretty', 'good', 'teams', 'to', 'get', 'here', ',', '"', 'Slocum', 'said', '.'],
            self.t.tokenize('"We beat some pretty good teams to get here," Slocum said.')
        )

    def test_brackets_quotes(self):
        """Test the word tokenizer on brackets and quotes."""
        self.assertEqual(
            ['Well', ',', 'we', 'could', 'n\'t', 'have', 'this', 'predictable', ',', 'cliche', '-', 'ridden', ',', '"', 'Touched', 'by', 'an', 'Angel', '"', '(', 'a', 'show', 'creator', 'John', 'Masius', 'worked', 'on', ')', 'wanna-be', 'if', 'she', 'did', 'n\'t', '.'],
            self.t.tokenize('Well, we couldn\'t have this predictable, cliche-ridden, "Touched by an Angel" (a show creator John Masius worked on) wanna-be if she didn\'t.')
        )

    def test_exclamation(self):
        """Test the word tokenizer on exclamation mark."""
        self.assertEqual(
            ['I', 'can', 'not', 'can', 'not', 'work', 'under', 'these', 'conditions', '!'],
            self.t.tokenize('I cannot cannot work under these conditions!')
        )

    def test_digit_comma(self):
        """Test the word tokenizer on commas within numbers."""
        self.assertEqual(
            ['The', 'company', 'spent', '$', '30,000,000', 'last', 'year', '.'],
            self.t.tokenize('The company spent $30,000,000 last year.')
        )

    def test_percentage(self):
        """Test the word tokenizer on percent sign."""
        self.assertEqual(
            ['The', 'company', 'spent', '40.75', '%', 'of', 'its', 'income', 'last', 'year', '.'],
            self.t.tokenize('The company spent 40.75% of its income last year.')
        )

    def test_colon_time(self):
        """Test the word tokenizer on colon between digits in a time."""
        self.assertEqual(
            ['He', 'arrived', 'at', '3', ':', '00', 'pm', '.'],
            self.t.tokenize('He arrived at 3:00 pm.')
        )

    def test_word_colon(self):
        """Test the word tokenizer on colon after word."""
        self.assertEqual(
            ['I', 'bought', 'these', 'items', ':', 'books', ',', 'pencils', ',', 'and', 'pens', '.'],
            self.t.tokenize('I bought these items: books, pencils, and pens.')
        )

    def test_digit_comma_space(self):
        """Test the word tokenizer on comma between digits with a space."""
        self.assertEqual(
            ['Though', 'there', 'were', '150', ',', '100', 'of', 'them', 'were', 'old', '.'],
            self.t.tokenize('Though there were 150, 100 of them were old.')
        )

    def test_digit_comma_multiple(self):
        """Test the word tokenizer on comma at end of digits."""
        self.assertEqual(
            ['There', 'were', '300,000', ',', 'but', 'that', 'was', "n't", 'enough', '.'],
            self.t.tokenize('There were 300,000, but that wasn\'t enough.')
        )

    def test_theyll(self):
        """Test the word tokenizer on the word they'll."""
        self.assertEqual(
            ['They', "'ll", 'save', 'and', 'invest', 'more', '.'],
            self.t.tokenize('They\'ll save and invest more.')
        )

    def test_bracket1(self):
        """Test the word tokenizer on sentence containing brackets."""
        self.assertEqual(
            ['For', 'a', 'few', 'weeks', '(', '>', '24', 'days', ')', '.'],
            self.t.tokenize('For a few weeks (>24 days).')
        )

    def test_bracket2(self):
        """Test the word tokenizer on sentence containing brackets."""
        self.assertEqual(
            ['Coumarin', '343', '(', 'C343', ')', 'was', 'added', 'to', 'the', 'mixture', '.'],
            self.t.tokenize('Coumarin 343 (C343) was added to the mixture.')
        )

    def test_bracket3(self):
        """Test the word tokenizer on sentence containing brackets."""
        self.assertEqual(
            ['(', 'Ka', ':', '1.42', '×', '10(10)', 'M-1', 'vs', '1.95', '±', '0.35', '×', '10(10)', 'M-1', ')', 'and', 'increased', '(', '9.9', 'vs', '3.7', '±', '0.4', 'fmol', ')'],
            self.t.tokenize('(Ka: 1.42×10(10) M-1 vs 1.95±0.35×10(10) M-1) and increased (9.9 vs 3.7±0.4 fmol)')
        )

    def test_bracket4(self):
        """Test the word tokenizer on sentence containing chemical name with commas and brackets."""
        self.assertEqual(
            ['Pd-C', 'hydrogenation', 'of', '3,21-diacetoxy-5', 'beta,19-cyclopregna-2,9(11)-diene-4,20-dione', '(', '10', ')'],
            self.t.tokenize('Pd-C hydrogenation of 3,21-diacetoxy-5 beta,19-cyclopregna-2,9(11)-diene-4,20-dione (10)')
        )

    def test_bracket5(self):
        """Test the word tokenizer on sentence containing chemical name with commas and brackets."""
        self.assertEqual(
            ['(2E)-3-[5-(2,3-Dimethoxy-6-methyl-1,4-benzoquinoyl)]-2-nonyl-2-propenoic', 'acid'],
            self.t.tokenize('(2E)-3-[5-(2,3-Dimethoxy-6-methyl-1,4-benzoquinoyl)]-2-nonyl-2-propenoic acid')
        )

    def test_bracket6(self):
        """Test the word tokenizer on sentence containing chemical name with prime and brackets."""
        self.assertEqual(
            ['(2-Amino-3\'-methoxyphenyl)oxanaphthalen-4-one', '(', 'PD', '98059', ')'],
            self.t.tokenize('(2-Amino-3\'-methoxyphenyl)oxanaphthalen-4-one (PD 98059)')
        )

    def test_bracket7(self):
        """Test the word tokenizer on sentence containing chemical name with brackets."""
        self.assertEqual(
            ['The', 'maximum', '(3H)-thymidine', 'incorporation', 'into', 'DNA', '.'],
            self.t.tokenize('The maximum (3H)-thymidine incorporation into DNA.')
        )

    def test_comma1(self):
        """Test the word tokenizer on sentence containing chemical name with commas."""
        self.assertEqual(
            ['With', 'hot', 'acetic', '+', 'p-toluenesulfonic', 'acid', '5', 'underwent', 'rearrangement', 'to', '12-acetoxy-11', 'beta,19-epoxypregn-5-ene-4,20-dione', '(', '8', ')', '.'],
            self.t.tokenize('With hot acetic + p-toluenesulfonic acid 5 underwent rearrangement to 12-acetoxy-11 beta,19-epoxypregn-5-ene-4,20-dione (8).')
        )

    def test_comma2(self):
        """Test the word tokenizer on sentence containing chemical name with commas."""
        self.assertEqual(
            ['N,N-Dimethylformamide', '(', 'DMF', ')', 'was', 'used', 'as', 'solvent', '.'],
            self.t.tokenize('N,N-Dimethylformamide (DMF) was used as solvent.')
        )

    def test_comma3(self):
        """Test the word tokenizer on sentence containing chemical name with commas."""
        self.assertEqual(
            ['1,2-bis(o-aminophenoxy)-ethane-N,N,N\',N\'-tetraacetic', 'acid'],
            self.t.tokenize('1,2-bis(o-aminophenoxy)-ethane-N,N,N\',N\'-tetraacetic acid')
        )

    def test_comma4(self):
        """Test the word tokenizer on sentence containing chemical name with commas."""
        self.assertEqual(
            ['N,N,N\',N\'-tetrakis', '[2-pyridylmethyl]ethylenediamine'],
            self.t.tokenize('N,N,N\',N\'-tetrakis [2-pyridylmethyl]ethylenediamine')
        )

    def test_comma5(self):
        """Test the word tokenizer on sentence containing chemical name with commas."""
        self.assertEqual(
            ['o,o\'-bismyristoyl', 'and', 'α,α\'-dipyridyl'],
            self.t.tokenize('o,o\'-bismyristoyl and α,α\'-dipyridyl')
        )

    def test_comma6(self):
        """Test the word tokenizer on sentence containing chemical name with a prime character."""
        self.assertEqual(
            ['The', 'characteristics', 'of', 'the', 'C(sp,sp2,sp3)', '–', 'H', '⋯', 'F', '–', 'C(sp,sp2,sp3)', 'intermolecular'],
            self.t.tokenize('The characteristics of the C(sp,sp2,sp3)–H⋯F–C(sp,sp2,sp3) intermolecular')
        )

    def test_prime(self):
        """Test the word tokenizer on sentence containing chemical name with a prime character."""
        self.assertEqual(
            ['N,N′-heteroaromatic', 'ancillary', 'ligands'],
            self.t.tokenize('N,N′-heteroaromatic ancillary ligands')
        )

    def test_square_bracket(self):
        """Test the word tokenizer on sentence containing chemical name with square brackets."""
        self.assertEqual(
            ['[Hg(dman)2]', ',', 'with', '[PdCl2(SEt2)2]', 'or', '[{PdCl2(PPh3)}2]'],
            self.t.tokenize('[Hg(dman)2], with [PdCl2(SEt2)2] or [{PdCl2(PPh3)}2]')
        )

    def test_curly_bracket(self):
        """Test the word tokenizer on sentence containing chemical name with curly brackets."""
        self.assertEqual(
            ['bimetallic', 'complexes', ',', '{[Mn(salpn)]2[Fe(CN)5NO]}n', '(', '1', ')', ',', '{[Mn(salpn)(CH3OH)]4[Mn(CN)5NO]}[C(CN)3]·3H2O', '(', '2', ')', ',', '{[Mn(dapsc)][Fe(CN)5NO]·0.5CH3OH·0.25H2O}n', '(', '3', ')', 'and', '{[Mn(salpn)(CH3OH)]4[Fe(CN)5NO]}(ClO4)2·4H2O', '(', '4', ')', ',', 'where', 'salpn2−', '=', 'N,N′-1,3-propylene-bis(salicylideneiminato)', 'dianion', 'and', 'dapsc', '=', '2,6-diacetylpyridine-bis(semicarbazone)'],
            self.t.tokenize('bimetallic complexes, {[Mn(salpn)]2[Fe(CN)5NO]}n (1), {[Mn(salpn)(CH3OH)]4[Mn(CN)5NO]}[C(CN)3]·3H2O (2), {[Mn(dapsc)][Fe(CN)5NO]·0.5CH3OH·0.25H2O}n (3) and {[Mn(salpn)(CH3OH)]4[Fe(CN)5NO]}(ClO4)2·4H2O (4), where salpn2− = N,N′-1,3-propylene-bis(salicylideneiminato) dianion and dapsc = 2,6-diacetylpyridine-bis(semicarbazone)')
        )

    def test_bracket_internal_split(self):
        """Test the word tokenizer on sentence containing token that should be split internally and from brackets."""
        self.assertEqual(
            ['Extensive', 'H', '-', 'bonding', '(', 'F', '⋯', 'H', '–', 'OH', ')', 'links', 'the', 'molecules'],
            self.t.tokenize('Extensive H-bonding (F⋯H–OH) links the molecules')
        )

    def test_colon_chem(self):
        """Test the word tokenizer on chemical name containing colon."""
        self.assertEqual(
            ['Group', '13', 'fluorides', ',', 'MF3·3H2O', '(', 'M', '=', 'Al', ',', 'Ga', 'or', 'In', ')', 'with', '2,2′:6′,2′′-terpyridyl', ',', '2,2′-bipyridyl', 'or', '1,10-phenanthroline', 'under', 'hydrothermal', 'conditions', '(', '180', '°', 'C', '/', '15', 'h', ')'],
            self.t.tokenize('Group 13 fluorides, MF3·3H2O (M = Al, Ga or In) with 2,2′:6′,2′′-terpyridyl, 2,2′-bipyridyl or 1,10-phenanthroline under hydrothermal conditions (180 °C/15 h)')
        )

    def test_positive_charge(self):
        """Test the word tokenizer on chemical name with trailing plus."""
        self.assertEqual(
            ['[⊂Me2N(CH2)2NMe(CH2)2]+', ',', 'with', 'fluorometallate', 'anions'],
            self.t.tokenize('[⊂Me2N(CH2)2NMe(CH2)2]+, with fluorometallate anions')
        )

    def test_bracket_subsequent(self):
        """Test the word tokenizer on chemical name with subsequent (not nested) brackets."""
        self.assertEqual(
            ['The', 'α-ammonium-acylchloride', 'salts', '[NH2(CH2)3CHC(O)Cl][WOCl5]', ',', '1a', ',', 'and', '[MeNH2CH2C(O)Cl][WOCl5]', ',', '1b', '.'],
            self.t.tokenize('The α-ammonium-acylchloride salts [NH2(CH2)3CHC(O)Cl][WOCl5], 1a, and [MeNH2CH2C(O)Cl][WOCl5], 1b.')
        )

    def test_bracket_subsequent2(self):
        """Test the word tokenizer on chemical name with subsequent (not nested) brackets."""
        self.assertEqual(
            ['(N,N)-thingy-(errrm)'],
            self.t.tokenize('(N,N)-thingy-(errrm)')
        )

    def test_bracket_range(self):
        """Test the word tokenizer on a bracketed range."""
        self.assertEqual(
            ['(', 'a', ')', '–', '(', 'c', ')', 'Some', 'things', '.'],
            self.t.tokenize('(a)–(c) Some things.')
        )

    def test_space_colon_digit(self):
        """Test the word tokenizer on an usual spacing of colon."""
        self.assertEqual(
            ['Intensity', 'ratio', 'of', '2', ':', '2', ':', '4', 'and', '54', ':', '18', '.'],
            self.t.tokenize('Intensity ratio of 2 :2 :4 and 54 : 18.')
        )

    def test_quote_apostrophe(self):
        """Test the word tokenizer when a quote is used as an apostrophe."""
        self.assertEqual(
            ['Alzheimer', '\u2019s', 'disease', '(', 'AD', ')', 'was', '‘', 'extremely', '’', 'hard', '‘', 'to', 'diagnose', '.', '’'],
            self.t.tokenize('Alzheimer’s disease (AD) was ‘extremely’ hard ‘to diagnose.’')
        )

    def test_quote_apostrophe2(self):
        """Test the word tokenizer when a quote is used as an apostrophe."""
        self.assertEqual(
            ['Alzheimer', '\u2019s', 'disease', '(', 'AD', ')', 'was', '‘', 'extremely', '’', 'hard', '‘', 'to', 'diagnose', '’', '.'],
            self.t.tokenize('Alzheimer’s disease (AD) was ‘extremely’ hard ‘to diagnose’.')
        )

    def test_quote_apostrophe3(self):
        """Test the word tokenizer when a quote is used as an apostrophe."""
        self.assertEqual(
            ['Alzheimer', '\u2019s', 'disease', '(', 'AD', ')', 'was', '‘', 'extremely', '’', 'hard', '‘', 'to', 'diagnose', '’', ',', 'at', 'the', 'time', '.'],
            self.t.tokenize('Alzheimer’s disease (AD) was ‘extremely’ hard ‘to diagnose’, at the time.')
        )

    def test_apostrophe_quote(self):
        """Test the word tokenizer when an apostrophe is used as a quote."""
        self.assertEqual(
            ["a", "Special", "Issue", "entitled", "'", "Cognitive", "Enhancers", "'", "."],
            self.t.tokenize("a Special Issue entitled 'Cognitive Enhancers'.")
        )

    def test_double_apostrophe_quote(self):
        """Test the word tokenizer when an apostrophe is used as a quote."""
        self.assertEqual(
            ["between", "the", "''", "Schiff", "base", "''", "and"],
            self.t.tokenize("between the ''Schiff base'' and")
        )

    def test_bracketed_url(self):
        """Test the word tokenizer on a bracketed URL."""
        self.assertEqual(
            ['QALIBRA', 'software', '(', 'www.qalibra.eu', ')'],
            self.t.tokenize('QALIBRA software (www.qalibra.eu)')
        )

    def test_symbols(self):
        self.assertEqual(['2', '+', '2', '=', '4'], self.t.tokenize('2+2=4'))
        self.assertEqual(['+', '4', '°', 'C'], self.t.tokenize('+4°C'))
        self.assertEqual(['(', 'H2A', '<-->', 'HA-', '+', 'H+', 'and', 'HA-', '<-->', 'A', '=', '+', 'H+', ')'], self.t.tokenize('(H2A <--> HA- + H+ and HA- <--> A= + H+)'))
        self.assertEqual(['[2+2+2]'], self.t.tokenize('[2+2+2]'))
        self.assertEqual(['95.5', '+/-', '0.2', '%'], self.t.tokenize('95.5 +/- 0.2%'))

    def test_sentence_end(self):
        self.assertEqual(['upon', 'addition', 'of', 'Ni(II)', ';'], self.t.tokenize('upon addition of Ni(II);'))
        self.assertEqual(['upon', 'addition', 'of', 'Ni(II)', '.'], self.t.tokenize('upon addition of Ni(II).'))
        self.assertEqual(['complexes', 'in', 'THF', '(', 'ii', ')', '.'], self.t.tokenize('complexes in THF (ii).'))
        self.assertEqual(['complexes', 'in', 'THF', '(', 'ii', ')', ','], self.t.tokenize('complexes in THF (ii),'))
        self.assertEqual(['measured', 'at', '303', 'K', '.'], self.t.tokenize('measured at 303 K.'))
        self.assertEqual(['Sentence', 'trails', 'off', '…'], self.t.tokenize('Sentence trails off…'))
        self.assertEqual(['Sentence', 'trails', 'off', '...'], self.t.tokenize('Sentence trails off...'))
        self.assertEqual(['in', 'the', 'AUC', '.'], self.t.tokenize('in the AUC.'))
        self.assertEqual(['for', 'lane', 'no.', '11', '.'], self.t.tokenize('for lane no. 11.'))
        self.assertEqual(['under', 'A.', 'M.', '1.5', 'illumination'], self.t.tokenize('under A. M. 1.5 illumination'))
        self.assertEqual(['space', 'group', 'P', '(', 'No.', '2', ')', '.'], self.t.tokenize('space group P (No. 2).'))

    def test_abbreviations(self):
        self.assertEqual(['(', 'ca.', '30', 'mL', ')'], self.t.tokenize('(ca. 30 mL)'))
        self.assertEqual(['Elements', ',', 'e.g.', 'calcium'], self.t.tokenize('Elements, e.g. calcium'))

    def test_more_brackets(self):
        self.assertEqual(['NaOH', '(', 'aq', ')'], self.t.tokenize('NaOH(aq)'))
        self.assertEqual(['HCl', '(', 'g', ')'], self.t.tokenize('HCl(g)'))
        self.assertEqual(['5', '(', 'g', ')'], self.t.tokenize('5(g)'))
        self.assertEqual(['a', ')', 'UV', '/', 'vis', 'spectrum', '.'], self.t.tokenize('a) UV/vis spectrum.'))
        self.assertEqual(['a', ')', 'UV', '-', 'vis', 'spectrum', '.'], self.t.tokenize('a) UV-vis spectrum.'))
        self.assertEqual(['(', 'c', ')', '–', '(', 'e', ')'], self.t.tokenize('(c)–(e)'))
        self.assertEqual(['THF', '(', 'i', ')', ',', 'toluene', '(', 'iii', ')'], self.t.tokenize('THF (i), toluene (iii)'))
        self.assertEqual(['buffer', '(', 'pH', '7.4', ')', '.'], self.t.tokenize('buffer (pH 7.4).'))

    def test_multihyphens(self):
        self.assertEqual(['---'], self.t.tokenize('---'))
        self.assertEqual(['–––'], self.t.tokenize('–––'))
        self.assertEqual(['———'], self.t.tokenize('———'))
        self.assertEqual(['−−−'], self.t.tokenize('−−−'))
        self.assertEqual(['--'], self.t.tokenize('--'))
        self.assertEqual(['––'], self.t.tokenize('––'))
        self.assertEqual(['——'], self.t.tokenize('——'))
        self.assertEqual(['−−'], self.t.tokenize('−−'))

    def test_tilde(self):
        self.assertEqual(['a', 'line', 'width', 'of', '\u223c', '3', 'Hz', '.'], self.t.tokenize('a line width of ∼3 Hz.'))
        self.assertEqual(['a', 'line', 'width', 'of', '~', '3', 'Hz', '.'], self.t.tokenize('a line width of ~3 Hz.'))

    def test_slashes(self):
        self.assertEqual(['methanol', '/', 'water'], self.t.tokenize('methanol/water'))
        self.assertEqual(['B3LYP', '/', '6-311G(d,p)'], self.t.tokenize('B3LYP/6-311G(d,p)'))

    def test_iron_states(self):
        self.assertEqual(['Fe(III)'], self.t.tokenize('Fe(III)'))
        self.assertEqual(['Fe(iii)'], self.t.tokenize('Fe(iii)'))
        self.assertEqual(['Fe(3+)'], self.t.tokenize('Fe(3+)'))
        self.assertEqual(['Fe(0)'], self.t.tokenize('Fe(0)'))

    def test_identifiers(self):
        self.assertEqual(['4CN'], self.t.tokenize('4CN'))
        self.assertEqual(['2a'], self.t.tokenize('2a'))

    def test_colons(self):
        self.assertEqual(['ethanol', ':', 'water'], self.t.tokenize('ethanol:water'))
        self.assertEqual(['1', ':', '2'], self.t.tokenize('1:2'))
        self.assertEqual(['1', ':', '2'], self.t.tokenize('1 : 2'))
        self.assertEqual(['(', 'foo', ')', ':', '(', 'bar', ')'], self.t.tokenize('(foo):(bar)'))
        self.assertEqual(['foo', ')', ':', '(', 'bar'], self.t.tokenize('foo):(bar'))
        self.assertEqual(['4:7,10:13-diepoxy[15]annulenone'], self.t.tokenize('4:7,10:13-diepoxy[15]annulenone'))
        self.assertEqual(['9-(5′,5-diphenyl[1,1′:3′,1′′:3′′,1:3,1′′′′-quinquephenyl]-5′′-diyl)-9H-carbazole'], self.t.tokenize('9-(5′,5-diphenyl[1,1′:3′,1′′:3′′,1:3,1′′′′-quinquephenyl]-5′′-diyl)-9H-carbazole'))
        self.assertEqual(['9,9′-(5′-phenyl[1,1′:3′,1′′-terphenyl]-3,5-diyl)bis-9H-carbazole'], self.t.tokenize('9,9′-(5′-phenyl[1,1′:3′,1′′-terphenyl]-3,5-diyl)bis-9H-carbazole'))

    def test_lambda(self):
        self.assertEqual(['lambda5-phosphane'], self.t.tokenize('lambda5-phosphane'))
        self.assertEqual(['λ5-phosphane'], self.t.tokenize('λ5-phosphane'))

    def test_chem_names(self):
        self.assertEqual(['Tetrahydro', 'furan', '(', 'THF', ')'], self.t.tokenize('Tetrahydro furan (THF)'))
        self.assertEqual(['(S)-alanine'], self.t.tokenize('(S)-alanine'))
        self.assertEqual(['D-glucose'], self.t.tokenize('D-glucose'))
        self.assertEqual(['spiro[4.5]decane'], self.t.tokenize('spiro[4.5]decane'))
        self.assertEqual(['β-D-Glucose'], self.t.tokenize('β-D-Glucose'))
        self.assertEqual(['L-alanyl-L-glutaminyl-L-arginyl-O-phosphono-L-seryl-L-alanyl-L-proline'],
                         self.t.tokenize('L-alanyl-L-glutaminyl-L-arginyl-O-phosphono-L-seryl-L-alanyl-L-proline'))
        self.assertEqual(['aluminium(3+)'], self.t.tokenize('aluminium(3+)'))
        self.assertEqual(['1-methyl-2-methylidene-cyclohexane'],
                         self.t.tokenize('1-methyl-2-methylidene-cyclohexane'))

    def test_rings(self):
        self.assertEqual(["2,2':6',2''-Terphenyl-1,1',1''-triol"],
                         self.t.tokenize("2,2':6',2''-Terphenyl-1,1',1''-triol"))
        self.assertEqual(["phenothiazino[3',4':5,6][1,4]oxazino[2,3-i]benzo[5,6][1,4]thiazino[3,2-c]phenoxazine"],
                         self.t.tokenize("phenothiazino[3',4':5,6][1,4]oxazino[2,3-i]benzo[5,6][1,4]thiazino[3,2-c]phenoxazine"))

    def test_saccharide(self):
        self.assertEqual(['beta-D-Glucopyranosyl-(1->4)-D-glucose'], self.t.tokenize('beta-D-Glucopyranosyl-(1->4)-D-glucose'))
        self.assertEqual(['α-D-Glucopyranosyl-(1→4)-β-D-glucopyranose'], self.t.tokenize('α-D-Glucopyranosyl-(1→4)-β-D-glucopyranose'))
        self.assertEqual(['(1→4)-β-D-Glucan'], self.t.tokenize('(1→4)-β-D-Glucan'))
        self.assertEqual(['((1→2)-α-D-galacto)-(1→4)-β-D-Glucan'], self.t.tokenize('((1→2)-α-D-galacto)-(1→4)-β-D-Glucan'))

    def test_polymer(self):
        self.assertEqual([u"poly(2,2'-diamino-5-hexadecylbiphenyl-3,3'-diyl)"],
                         self.t.tokenize(u"poly(2,2'-diamino-5-hexadecylbiphenyl-3,3'-diyl)"))

    def test_operators(self):
        self.assertEqual(['J', '=', '8.8'], self.t.tokenize('J=8.8'))
        self.assertEqual(['CH2', '=', 'CH2'], self.t.tokenize('CH2=CH2'))
        self.assertEqual(['mL', '×', '3'], self.t.tokenize('mL×3'))
        self.assertEqual(['3', '×'], self.t.tokenize('3×'))
        self.assertEqual(['×', '3'], self.t.tokenize('×3'))
        self.assertEqual(['15', '÷', '3'], self.t.tokenize('15÷3'))
        self.assertEqual(['5', '+', '3'], self.t.tokenize('5+3'))
        self.assertEqual(['ESI+'], self.t.tokenize('ESI+'))
        self.assertEqual(['Ce3+'], self.t.tokenize('Ce3+'))

    def test_stereo(self):
        self.assertEqual(['(+)-chiraline'], self.t.tokenize('(+)-chiraline'))
        self.assertEqual(['(-)-chiraline'], self.t.tokenize('(-)-chiraline'))
        self.assertEqual(['(+-)-chiraline'], self.t.tokenize('(+-)-chiraline'))  # \u002d Hyphen-minus
        self.assertEqual(['(+−)-chiraline'], self.t.tokenize('(+−)-chiraline'))  # \u2212 Minus
        self.assertEqual(['(+/-)-chiraline'], self.t.tokenize('(+/-)-chiraline'))  # \u002d Hyphen-minus
        self.assertEqual(['(+/−)-chiraline'], self.t.tokenize('(+/−)-chiraline'))  # \u2212 Minus
        self.assertEqual(['(±)-chiraline'], self.t.tokenize('(±)-chiraline'))

    def test_hyphen_twice(self):
        self.assertEqual(['cytoplasmic', '-', 'to', '-', 'nuclear'], self.t.tokenize('cytoplasmic-to-nuclear'))
        self.assertEqual(['layer', '-', 'by', '-', 'layer'], self.t.tokenize('layer-by-layer'))
        self.assertEqual(['end', '-', 'of', '-', 'phase'], self.t.tokenize('end-of-phase'))
        self.assertEqual(['oil', '-', 'in', '-', 'water'], self.t.tokenize('oil-in-water'))
        self.assertEqual(['nucleation', '-', 'and', '-', 'growth'], self.t.tokenize('nucleation-and-growth'))
        self.assertEqual(['State', '-', 'of', '-', 'the', '-', 'art'], self.t.tokenize('State-of-the-art'))

    def test_nmr_types(self):
        self.assertEqual(['(13)C', '-', 'NMR'], self.t.tokenize('(13)C-NMR'))
        self.assertEqual(['(1)H', '-', 'NMR'], self.t.tokenize('(1)H-NMR'))
        self.assertEqual(['(31)P', '-', 'NMR'], self.t.tokenize('(31)P-NMR'))
        self.assertEqual(['(19)F', '-', 'NMR'], self.t.tokenize('(19)F-NMR'))
        self.assertEqual(['1H', '-', 'NMR'], self.t.tokenize('1H-NMR'))
        self.assertEqual(['31P', '-', 'NMR'], self.t.tokenize('31P-NMR'))
        self.assertEqual(['19F', '-', 'NMR'], self.t.tokenize('19F-NMR'))
        self.assertEqual(['proton', '-', 'NMR'], self.t.tokenize('proton-NMR'))

    def test_bracket_hyphen(self):
        self.assertEqual(['(LBD)-linked'], self.t.tokenize('(LBD)-linked'))
        self.assertEqual(['Fe(IV)', '-', 'oxo-mediated'], self.t.tokenize('Fe(IV)-oxo-mediated'))
        self.assertEqual(['Fe(IV)', '-', 'mediated'], self.t.tokenize('Fe(IV)-mediated'))
        self.assertEqual(['T-bet(-/-)'], self.t.tokenize('T-bet(-/-)'))
        self.assertEqual(['(', 'nano', 'LC', '/', 'nano-ESI-IT-MS', ')'], self.t.tokenize('(nano LC/nano-ESI-IT-MS)'))

    def test_hyphen_nosplit(self):
        self.assertEqual(['1,4-addition'], self.t.tokenize('1,4-addition'))
        self.assertEqual(['1,3\'-substituted'], self.t.tokenize('1,3\'-substituted'))
        self.assertEqual(['3,3′-dianisyl-substituted'], self.t.tokenize('3,3′-dianisyl-substituted'))
        self.assertEqual(['α-substituted'], self.t.tokenize('α-substituted'))
        self.assertEqual(['meta-substituted'], self.t.tokenize('meta-substituted'))
        self.assertEqual(['poly-zwitterion'], self.t.tokenize('poly-zwitterion'))
        self.assertEqual(['1,2-zwitterion'], self.t.tokenize('1,2-zwitterion'))

    def test_element_hyphen(self):
        self.assertEqual(['Fe', '-', 'containing'], self.t.tokenize('Fe-containing'))
        self.assertEqual(['C', '-', 'terminal'], self.t.tokenize('C-terminal'))
        self.assertEqual(['Li', '-', 'ions'], self.t.tokenize('Li-ions'))

    def test_hyphen_split(self):
        self.assertEqual(['hydrocarbon', '-', 'based'], self.t.tokenize('hydrocarbon-based'))
        self.assertEqual(['methicillin', '-', 'resistant'], self.t.tokenize('methicillin-resistant'))
        self.assertEqual(['methicillin', '-', 'resistant', ','], self.t.tokenize('methicillin-resistant,'))
        self.assertEqual(['HPMA', '-', 'based'], self.t.tokenize('HPMA-based'))
        self.assertEqual(['HPMA', '-', 'based', ')'], self.t.tokenize('HPMA-based)'))

    def test_bracket_chem2(self):
        """Test the word tokenizer on chemical name containing brackets."""
        self.assertEqual(
            ['(-)-(5R,8S,8aS)-8-methyl-5-pentyloctahydroindolizine', '(', '8-epi-indolizidine', '209B', ')', '9', 'in', '74', '%', 'yield', '.'],
            self.t.tokenize('(-)-(5R,8S,8aS)-8-methyl-5-pentyloctahydroindolizine (8-epi-indolizidine 209B) 9 in 74% yield.')
        )

    def test_bracket_chem_identifier(self):
        """Test the word tokenizer on chemical identifier with stereo brackets."""
        self.assertEqual(
            ['produced', 'the', 'thiolactam', '(+)-27', 'in', '92', '%', 'yield', '.'],
            self.t.tokenize('produced the thiolactam (+)-27 in 92% yield.')
        )

    def test_minus_hyphen(self):
        """Test the word tokenizer on a minus used as a hyphen."""
        self.assertEqual(
            ['The', 'dose', '−', 'response', 'curve', '.'],
            self.t.tokenize('The dose−response curve.')
        )

    def test_abbreviation_sentence_end(self):
        """Test the word tokenizer on sentence with abbreviation at the end."""
        self.assertEqual(['Chemical', 'Company', 'Ltd.'], self.t.tokenize('Chemical Company Ltd.'))
        self.assertEqual(['Studies', 'in', 'the', 'U.S.'], self.t.tokenize('Studies in the U.S.'))
        self.assertEqual(['the', 'mean', '±', 'S.D.'], self.t.tokenize('the mean ± S.D.'))
        self.assertEqual(['in', 'a', 'beaker', 'at', 'r.t.'], self.t.tokenize('in a beaker at r.t.'))
        self.assertEqual(['Whitaker', 'et', 'al.'], self.t.tokenize('Whitaker et al.'))

    def test_trademarks(self):
        self.assertEqual(['CML', '(', 'TM', ')'], self.t.tokenize('CML(TM)'))
        self.assertEqual(['Apple', '(', 'R', ')'], self.t.tokenize('Apple(R)'))
        self.assertEqual(['IR3535', '(', '®', ')'], self.t.tokenize('IR3535(®)'))
        self.assertEqual(['IR3535', '(', '™', ')'], self.t.tokenize('IR3535(™)'))
        self.assertEqual(['IR3535', '(', 'TM', ')'], self.t.tokenize('IR3535(TM)'))
        self.assertEqual(['IR3535', '(', 'R', ')'], self.t.tokenize('IR3535(R)'))
        self.assertEqual(['IR3535', '®'], self.t.tokenize('IR3535®'))
        self.assertEqual(['IR3535', '™'], self.t.tokenize('IR3535™'))

    def test_ms(self):
        self.assertEqual(['[M+H]+', '1523.86', ',', '[M+2H]2+', '762.43', ',', '[M+3H]3+', '508.62.', 'Observed', ':', '[M+H]+', '1523.20', ',', '[M+2H]2+', '762.45', ',', '[M+3H]3+', '508.70', '.'], self.t.tokenize('[M+H]+ 1523.86, [M+2H]2+ 762.43, [M+3H]3+ 508.62. Observed: [M+H]+ 1523.20, [M+2H]2+ 762.45, [M+3H]3+ 508.70.'))
        # This isn't ideal but can't see any alternative apart from super fine tokenization
        self.assertEqual(['527.3596', '[', 'M', '+', 'H]+', ',', 'C30H47N4O4+'], self.t.tokenize('527.3596 [M + H]+, C30H47N4O4+'))

    def test_quantities(self):
        self.assertEqual(['contamination', 'of', '2', '%', 'Cl2'], self.t.tokenize('contamination of 2% Cl2'))
        self.assertEqual(['Placed', 'at', 'a', 'distance', 'of', '7.2', 'cm', '.'], self.t.tokenize('Placed at a distance of 7.2cm.'))
        self.assertEqual(['Addition', 'of', '~', '1.8', 'mg', 'of', 'CaCO3', '.'], self.t.tokenize('Addition of ~1.8mg of CaCO3.'))
        self.assertEqual(['Recorded', 'in', 'HCl', '(', 'pH', '2', ')', '.'], self.t.tokenize('Recorded in HCl (pH2).'))
        self.assertEqual(['Experienced', 'a', 'pressure', 'of', '160', 'kPa', '.'], self.t.tokenize('Experienced a pressure of 160kPa.'))
        self.assertEqual(['Brought', 'to', 'pH', '10.5', ',', 'gradually', '.'], self.t.tokenize('Brought to pH10.5, gradually.'))
        self.assertEqual(['A', 'volume', 'of', '24', 'cm3', 'was', 'drained', '.'], self.t.tokenize('A volume of 24cm3 was drained.'))
        self.assertEqual(['2', 'M', 'H2SO4', 'was', 'heated', '.'], self.t.tokenize('2M H2SO4 was heated.'))
        self.assertEqual(['The', 'spectrum', 'was', 'recorded', 'at', '10', '°', 'C'], self.t.tokenize('The spectrum was recorded at 10° C'))
        self.assertEqual(['The', 'spectrum', 'was', 'recorded', 'at', '10', '°', 'C'], self.t.tokenize('The spectrum was recorded at 10°C'))
        self.assertEqual(['The', 'spectrum', 'was', 'recorded', 'at', '10', '°', 'C'], self.t.tokenize('The spectrum was recorded at 10 °C'))
        self.assertEqual(['Added', '3.5', 'g', 'and', 'stirred', 'for', '5.5', 's', '.'], self.t.tokenize('Added 3.5g and stirred for 5.5s.'))
        self.assertEqual(['and', '≈', '90', '°'], self.t.tokenize('and ≈90°'))
        self.assertEqual(['B3LYP', '/', '6-31g(d)'], self.t.tokenize('B3LYP/6-31g(d)'))
        self.assertEqual(['N', '1s', 'spectra'], self.t.tokenize('N 1s spectra'))
        self.assertEqual(['In', 'the', '1980s', 'there', 'was'], self.t.tokenize('In the 1980s there was'))
        self.assertEqual(['Produced', 'compounds', '3g', ',', '3l', ',', '3m', 'and', '3n'], self.t.tokenize('Produced compounds 3g, 3l, 3m and 3n'))
        self.assertEqual(['9.66', '(', 'd', ',', '1H', ',', '3J', '=', '5.4', 'Hz', ',', 'H15', ')'], self.t.tokenize('9.66 (d, 1H, 3J = 5.4Hz, H15)'))
        self.assertEqual(['greater', 'than', '3', '×', '10-11', 'mol', 'kg-1', 'or', '2', '×', '10-5', 'mol', 'kg-1'], self.t.tokenize('greater than 3 × 10-11mol kg-1 or 2 × 10-5mol kg-1'))

    def test_linesymbols(self):
        self.assertEqual(['N', '(', '■', ')', ',', 'C2', '(', '●', ')', ',', 'C3', '(', '▲', ')'],
                         self.t.tokenize('N(■), C2(●), C3(▲)'))
        self.assertEqual(['benzaldehyde', '(', '○', ')'], self.t.tokenize('benzaldehyde (○)'))
        self.assertEqual(['6', '(', '--', ')', ',', '1', '(', '----', ')', 'and', '3', '(', '·····', ')'],
                         self.t.tokenize('6 (--), 1 (----) and 3 (·····)'))
        self.assertEqual(['6', '(', '--', ')', ',', '1', '(', '----', ')', 'and', '3', '(', '·····', ')'],
                         self.t.tokenize('6 (--), 1 (----) and 3 (·····)'))

    def test_bracket_chems(self):
        self.assertEqual(['molecules', 'of', 'the', '[NiL2]', 'complex'], self.t.tokenize('molecules of the [NiL2] complex'))
        self.assertEqual(['[Et3NBz][FeIIICl4]'], self.t.tokenize('[Et3NBz][FeIIICl4]'))
        self.assertEqual(['[2PA-Mmim][Tf2N]'], self.t.tokenize('[2PA-Mmim][Tf2N]'))
        self.assertEqual(['[H2O]', '≈', '3000', 'ppm'], self.t.tokenize('[H2O] ≈ 3000 ppm'))
        self.assertEqual(['(', '[Cu+]', '/', '[L]', '=', '3', ')'], self.t.tokenize('([Cu+]/[L] = 3)'))
        self.assertEqual(['(Ph3PO)(Ph3POH)(HSO4)'], self.t.tokenize('(Ph3PO)(Ph3POH)(HSO4)'))
        self.assertEqual(['(', 'iron(III)'], self.t.tokenize('(iron(III)'))

    def test_chem_formula(self):
        self.assertEqual(['(C2H5)4N'], self.t.tokenize('(C2H5)4N'))
        self.assertEqual(['(C2H5)4N'], self.t.tokenize('(C2H5)4N'))
        self.assertEqual(['monomer', '28M-Py2'], self.t.tokenize('monomer 28M-Py2'))
        self.assertEqual(['monomer', '28M-Py2'], self.t.tokenize('monomer 28M-Py2'))
        self.assertEqual(['ratio', 'Ag+', '/', 'nucleoside', 'of', '3', ':', '1'], self.t.tokenize('ratio Ag+/nucleoside of 3:1'))
        self.assertEqual(['[Al(H2L)n]3-'], self.t.tokenize('[Al(H2L)n]3-'))
        self.assertEqual(['[Fe(CN)5(NO)]2-'], self.t.tokenize('[Fe(CN)5(NO)]2-'))
        self.assertEqual(['[Fe(CN)5(NO)]2−'], self.t.tokenize('[Fe(CN)5(NO)]2−'))

    def test_deuterated(self):
        self.assertEqual(['acetone-d6'], self.t.tokenize('acetone-d6'))
        self.assertEqual(['chloroform-d'], self.t.tokenize('chloroform-d'))
        self.assertEqual(['d8-THF'], self.t.tokenize('d8-THF'))
        self.assertEqual(['THF-d8'], self.t.tokenize('THF-d8'))
        self.assertEqual(['d6-DMSO'], self.t.tokenize('d6-DMSO'))
        self.assertEqual(['DMSO-d6'], self.t.tokenize('DMSO-d6'))

    def test_reagents_list(self):
        """Test the word tokenizer on a reagents list."""
        self.assertEqual(
            ['Reagents', ':', '(', 'i', ')', 'H2', '(', '7', 'atm', ')', ',', '10', '%', 'Pd', '/', 'C', ',', 'AcOH', ',', 'rt', ';', '(', 'ii', ')', 'Cl(CH2)3COCl', ',', 'NaOEt', '(', 'cat.', ')', ',', 'CHCl3', ',', 'reflux', ';', '(', 'iii', ')', 'Lawesson', '\'s', 'reagent', ',', '110', '°', 'C', ';'],
            self.t.tokenize('Reagents: (i) H2 (7 atm), 10% Pd/C, AcOH, rt; (ii) Cl(CH2)3COCl, NaOEt (cat.), CHCl3, reflux; (iii) Lawesson\'s reagent, 110°C;')
        )

    def test_abbreviation_definition(self):
        """Test the word tokenizer on chemical abbreviation definition."""
        self.assertEqual(
            ['(', 'ADDP', ':', "1,1'-(azodicarbonyl)dipiperidine", ')'],
            self.t.tokenize('(ADDP: 1,1\'-(azodicarbonyl)dipiperidine)')
        )

    def test_nmr_whitespace_error(self):
        """Test the word tokenizer on NMR isotope missing preceding whitespace."""
        self.assertEqual(['726.1520', '.', '1H', 'NMR'], self.t.tokenize('726.1520.1H NMR'))
        self.assertEqual(['intermediate', '.', '1H', u'NMR'], self.t.tokenize('intermediate.1H NMR'))

    def test_ir_whitespace_error(self):
        """Test things like IR(KBr)."""
        self.assertEqual(['IR', '(', 'KBr', ')'], self.t.tokenize('IR(KBr)'))

    def test_bracket_whitespace_error(self):
        """Test the word tokenizer on bracket whitespace error."""
        self.assertEqual(['7.95', '(', 's', ',', '4H', ')'], self.t.tokenize('7.95(s, 4H)'))
        self.assertEqual(['In', 'Fig.', '5', '(', 'a', ',', 'b', ')'], self.t.tokenize('In Fig. 5(a, b)'))

    def test_quote_colon(self):
        """Test the word tokenizer quote followed by colon followed by digit (IndexError bugfix)."""
        self.assertEqual(['\'', ':', '1'], self.t.tokenize('\':1'))

    def test_chemtext_sentence(self):
        """Test tokenization through the Text and Sentence API."""
        t = Text('Hi, my name is Matt. What is your name?')
        self.assertEqual(
            [['Hi', ',', 'my', 'name', 'is', 'Matt', '.'], ['What', 'is', 'your', 'name', '?']],
            [sent.raw_tokens for sent in t.sentences]
        )

    def test_chemtext_sentence2(self):
        """Test tokenization through the ChemText and Sentence API."""
        t = Text('(Ka: 1.42×10(10) M-1 vs 1.95±0.35×10(10) M-1) and increased (9.9 vs 3.7±0.4 fmol)')
        self.assertEqual(
            [['(', 'Ka', ':', '1.42', '×', '10(10)', 'M-1', 'vs', '1.95', '±', '0.35', '×', '10(10)', 'M-1', ')', 'and', 'increased', '(', '9.9', 'vs', '3.7', '±', '0.4', 'fmol', ')']],
            [sent.raw_tokens for sent in t.sentences]
        )


class TestFineWordTokenizer(unittest.TestCase):
    """Test the fine word tokenizer."""

    maxDiff = None

    def setUp(self):
        self.t = FineWordTokenizer()

    def test_final_full_stop(self):
        """Test the word tokenizer splits off final full stop only."""
        self.assertEqual(
            ['This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.'],
            self.t.tokenize('This is Mr. Hoppy\'s tortoise.')
        )

    def test_full_stop_following(self):
        """Test the word tokenizer splits off final full stop if followed by brackets or quotes."""
        self.assertEqual(
            ['(', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.', ')'],
            self.t.tokenize('(This is Mr. Hoppy\'s tortoise.)')
        )
        self.assertEqual(
            ['"', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tortoise', '.', '"'],
            self.t.tokenize('"This is Mr. Hoppy\'s tortoise."')
        )
        self.assertEqual(
            ['"', 'This', 'is', 'Mr.', 'Hoppy', '\'s', 'tort.oise', '.', '"'],
            self.t.tokenize('"This is Mr. Hoppy\'s tort.oise."')
        )

    def test_dollar(self):
        """Test the word tokenizer on dollar symbol."""
        self.assertEqual(
            ['On', 'a', '$', '50,000', 'mortgage', 'of', '30', 'years', 'at', '8', 'percent', ',', 'the', 'monthly', 'payment', 'would', 'be', '$', '366.88', '.'],
            self.t.tokenize('On a $50,000 mortgage of 30 years at 8 percent, the monthly payment would be $366.88.')
        )

    def test_quote(self):
        """Test the word tokenizer on quotes."""
        self.assertEqual(
            ['"', 'We', 'beat', 'some', 'pretty', 'good', 'teams', 'to', 'get', 'here', ',', '"', 'Slocum', 'said', '.'],
            self.t.tokenize('"We beat some pretty good teams to get here," Slocum said.')
        )

    def test_brackets_quotes(self):
        """Test the word tokenizer on brackets and quotes."""
        self.assertEqual(
            ['Well', ',', 'we', 'could', 'n\'t', 'have', 'this', 'predictable', ',', 'cliche', '-', 'ridden', ',', '"', 'Touched', 'by', 'an', 'Angel', '"', '(', 'a', 'show', 'creator', 'John', 'Masius', 'worked', 'on', ')', 'wan', 'na', '-', 'be', 'if', 'she', 'did', 'n\'t', '.'],
            self.t.tokenize('Well, we couldn\'t have this predictable, cliche-ridden, "Touched by an Angel" (a show creator John Masius worked on) wanna-be if she didn\'t.')
        )

    def test_exclamation(self):
        """Test the word tokenizer on exclamation mark."""
        self.assertEqual(
            ['I', 'can', 'not', 'can', 'not', 'work', 'under', 'these', 'conditions', '!'],
            self.t.tokenize('I cannot cannot work under these conditions!')
        )

    def test_digit_comma(self):
        """Test the word tokenizer on commas within numbers."""
        self.assertEqual(
            ['The', 'company', 'spent', '$', '30,000,000', 'last', 'year', '.'],
            self.t.tokenize('The company spent $30,000,000 last year.')
        )

    def test_percentage(self):
        """Test the word tokenizer on percent sign."""
        self.assertEqual(
            ['The', 'company', 'spent', '40.75', '%', 'of', 'its', 'income', 'last', 'year', '.'],
            self.t.tokenize('The company spent 40.75% of its income last year.')
        )

    def test_colon_time(self):
        """Test the word tokenizer on colon between digits in a time."""
        self.assertEqual(
            ['He', 'arrived', 'at', '3', ':', '00', 'pm', '.'],
            self.t.tokenize('He arrived at 3:00 pm.')
        )

    def test_word_colon(self):
        """Test the word tokenizer on colon after word."""
        self.assertEqual(
            ['I', 'bought', 'these', 'items', ':', 'books', ',', 'pencils', ',', 'and', 'pens', '.'],
            self.t.tokenize('I bought these items: books, pencils, and pens.')
        )

    def test_digit_comma_space(self):
        """Test the word tokenizer on comma between digits with a space."""
        self.assertEqual(
            ['Though', 'there', 'were', '150', ',', '100', 'of', 'them', 'were', 'old', '.'],
            self.t.tokenize('Though there were 150, 100 of them were old.')
        )

    def test_digit_comma_multiple(self):
        """Test the word tokenizer on comma at end of digits."""
        self.assertEqual(
            ['There', 'were', '300,000', ',', 'but', 'that', 'was', "n't", 'enough', '.'],
            self.t.tokenize('There were 300,000, but that wasn\'t enough.')
        )

    def test_theyll(self):
        """Test the word tokenizer on the word they'll."""
        self.assertEqual(
            ['They', "'ll", 'save', 'and', 'invest', 'more', '.'],
            self.t.tokenize('They\'ll save and invest more.')
        )

    def test_bracket1(self):
        """Test the word tokenizer on sentence containing brackets."""
        self.assertEqual(
            ['For', 'a', 'few', 'weeks', '(', '>', '24', 'days', ')', '.'],
            self.t.tokenize('For a few weeks (>24 days).')
        )

    def test_bracket_chems(self):
        self.assertEqual(['molecules', 'of', 'the', '[', 'NiL2', ']', 'complex'], self.t.tokenize('molecules of the [NiL2] complex'))
        self.assertEqual(['[', 'Et3NBz', ']', '[', 'FeIIICl4', ']'], self.t.tokenize('[Et3NBz][FeIIICl4]'))
        self.assertEqual(['[', '2PA', '-', 'Mmim', ']', '[', 'Tf2N', ']'], self.t.tokenize('[2PA-Mmim][Tf2N]'))
        self.assertEqual(['[', 'H2O', ']', '≈', '3000', 'ppm'], self.t.tokenize('[H2O] ≈ 3000 ppm'))
        self.assertEqual(['(', '[', 'Cu', '+', ']', '/', '[', 'L', ']', '=', '3', ')'], self.t.tokenize('([Cu+]/[L] = 3)'))
        self.assertEqual(['(', 'Ph3PO', ')', '(', 'Ph3POH', ')', '(', 'HSO4', ')'], self.t.tokenize('(Ph3PO)(Ph3POH)(HSO4)'))
        self.assertEqual(['(', 'iron', '(', 'III', ')'], self.t.tokenize('(iron(III)'))

    def test_chem_formula(self):
        self.assertEqual(['(', 'C2H5', ')', '4N'], self.t.tokenize('(C2H5)4N'))
        self.assertEqual(['(', 'C2H5', ')', '4N'], self.t.tokenize('(C2H5)4N'))
        self.assertEqual(['monomer', '28M', '-', 'Py2'], self.t.tokenize('monomer 28M-Py2'))
        self.assertEqual(['monomer', '28M', '-', 'Py2'], self.t.tokenize('monomer 28M-Py2'))
        self.assertEqual(['ratio', 'Ag', '+', '/', 'nucleoside', 'of', '3', ':', '1'], self.t.tokenize('ratio Ag+/nucleoside of 3:1'))
        self.assertEqual(['[', 'Al', '(', 'H2L', ')', 'n', ']', '3', '-'], self.t.tokenize('[Al(H2L)n]3-'))
        self.assertEqual(['[', 'Fe', '(', 'CN', ')', '5', '(', 'NO', ')', ']', '2', '-'], self.t.tokenize('[Fe(CN)5(NO)]2-'))
        self.assertEqual(['[', 'Fe', '(', 'CN', ')', '5', '(', 'NO', ')', ']', '2', '−'], self.t.tokenize('[Fe(CN)5(NO)]2−'))

    def test_chem_names(self):
        self.assertEqual(['Tetrahydro', 'furan', '(', 'THF', ')'], self.t.tokenize('Tetrahydro furan (THF)'))
        self.assertEqual(['(', 'S', ')', '-', 'alanine'], self.t.tokenize('(S)-alanine'))
        self.assertEqual(['D', '-', 'glucose'], self.t.tokenize('D-glucose'))
        self.assertEqual(['spiro', '[', '4.5', ']', 'decane'], self.t.tokenize('spiro[4.5]decane'))
        self.assertEqual(['β', '-', 'D', '-', 'Glucose'], self.t.tokenize('β-D-Glucose'))
        self.assertEqual(['L', '-', 'alanyl', '-', 'L', '-', 'glutaminyl', '-', 'L', '-', 'arginyl', '-', 'O', '-', 'phosphono', '-', 'L', '-', 'seryl', '-', 'L', '-', 'alanyl', '-', 'L', '-', 'proline'],
                         self.t.tokenize('L-alanyl-L-glutaminyl-L-arginyl-O-phosphono-L-seryl-L-alanyl-L-proline'))
        self.assertEqual(['aluminium', '(', '3', '+', ')'], self.t.tokenize('aluminium(3+)'))
        self.assertEqual(['1', '-', 'methyl', '-', '2', '-', 'methylidene', '-', 'cyclohexane'],
                         self.t.tokenize('1-methyl-2-methylidene-cyclohexane'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
batterydataextractor.reader.elsevier

~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Elsevier XML reader
"""
from ..scrape.clean import clean, Cleaner
from ..scrape.elsevier import fix_elsevier_xml_whitespace, els_xml_whitespace, els_clean_abstract
from ..doc.meta import MetaData
from .markup import XmlReader
from lxml import etree

# XML stripper that removes the tags around numbers in chemical formulas
strip_els_xml = Cleaner(strip_xpath='.//ce:inf | .//ce:italic | .//ce:bold | .//ce:formula | .//mml:* | .//ce:sup'
                                    '| .//ce:cross-ref | .//ce:cross-refs | .//ja:math ',
                        kill_xpath='.//ce:cross-ref//ce:sup | .//ce:table//ce:sup | .//ce:float-anchor| .//ce:hsp'
                                   '| .//ja:math ')


class ElsevierXmlReader(XmlReader):
    """Reader for Elsevier XML documents."""

    cleaners = [clean, fix_elsevier_xml_whitespace, els_xml_whitespace, els_clean_abstract, strip_els_xml]

    etree.FunctionNamespace("http://www.elsevier.com/xml/svapi/article/dtd").prefix = 'default'
    etree.FunctionNamespace("http://www.elsevier.com/xml/bk/dtd").prefix = 'bk'
    etree.FunctionNamespace("http://www.elsevier.com/xml/common/cals/dtd").prefix = 'cals'
    etree.FunctionNamespace("http://www.elsevier.com/xml/common/dtd").prefix = 'ce'
    etree.FunctionNamespace("http://www.elsevier.com/xml/ja/dtd").prefix = 'ja'
    etree.FunctionNamespace("http://www.w3.org/1998/Math/MathML").prefix = 'mml'
    etree.FunctionNamespace("http://www.elsevier.com/xml/common/struct-aff/dtd").prefix = 'sa'
    etree.FunctionNamespace("http://www.elsevier.com/xml/common/struct-bib/dtd").prefix = 'sb'
    etree.FunctionNamespace("http://www.elsevier.com/xml/common/table/dtd").prefix = 'tb'
    etree.FunctionNamespace("http://www.w3.org/1999/xlink").prefix = 'xlink'
    etree.FunctionNamespace("http://www.elsevier.com/xml/xocs/dtd").prefix = 'xocs'
    etree.FunctionNamespace("http://purl.org/dc/elements/1.1/").prefix = 'dc'
    etree.FunctionNamespace("http://purl.org/dc/terms/").prefix = 'dcterms'
    etree.FunctionNamespace("http://prismstandard.org/namespaces/basic/2.0/").prefix = 'prism'
    etree.FunctionNamespace("http://www.w3.org/2001/XMLSchema-instance").prefix = 'xsi'

    root_css = 'default|full-text-retrieval-response'
    title_css = 'dc|title, ce|title'
    heading_css = 'ce|section-title'
    abstract_css = 'ce|abstract'
    table_css = 'ce|table'
    metadata_css = 'xocs|meta'
    metadata_title_css = 'xocs|normalized-article-title'
    metadata_author_css = 'xocs|normalized-first-auth-surname'
    metadata_journal_css = 'xocs|srctitle'
    metadata_volume_css = 'xocs|vol-first, xocs|volume-list xocs|volume'
    metadata_issue_css = 'xocs|issns xocs|issn-primary-formatted'
    metadata_publisher_css = 'xocs|copyright-line'
    metadata_date_css = 'xocs|available-online-date, xocs|orig-load-date'
    metadata_firstpage_css = 'xocs|first-fp'
    metadata_lastpage_css = 'xocs|last-lp'
    metadata_doi_css = 'xocs|doi, xocs|eii'
    metadata_pii_css = 'xocs|pii-unformatted'
    reference_css = 'ce|cross-refs, ce|cross-ref'
    citation_css = 'ce|bib-reference'
    ignore_css = 'ce|acknowledgment, ce|correspondence, ce|author, ce|doi, ja|jid, ja|aid, ce|pii, ' \
                 'xocs|oa-sponsor-type, xocs|open-access, default|openaccess, ce|article-number,'\
                 'default|openaccessArticle, dc|format, dc|creator, dc|identifier,'\
                 'default|eid, default|pii, xocs|ref-info, default|scopus-eid, '\
                 'xocs|normalized-srctitle, ' \
                 'xocs|eid, xocs|hub-eid, xocs|normalized-first-auth-surname,' \
                 'xocs|normalized-first-auth-initial, xocs|refkeys,' \
                 'xocs|attachment-eid, xocs|attachment-type, mml|math, mml|mrow, ' \
                 'ja|jid, ce|given-name, ce|surname, ce|affiliation, ce|label, ' \
                 'ce|grant-sponsor, ce|grant-number, prism|copyright,' \
                 'xocs|pii-unformatted, xocs|ucs-locator, ce|copyright,' \
                 'prism|publisher, prism|*, xocs|copyright-line, xocs|cp-notice,' \
                 'dc|description, ce|table, ce|figure, default|coredata, default|objects, default|scopus-id, ' \
                 'ce|nomenclature, ce|formula, ce|conflict-of-interest, ce|acknowledgements, default|aid'

    url_prefix = 'https://sciencedirect.com/science/article/pii/'

    def detect(self, fstring, fname=None):
        """Elsevier document detection based on string found in xml"""
        if fname and not fname.endswith('.xml'):
            return False
        if b'xmlns="http://www.elsevier.com/xml/svapi/article/dtd"' in fstring:
            return True
        return False

    def _parse_metadata(self, el, refs, specials):
        title = self._css(self.metadata_title_css, el)
        authors = self._css(self.metadata_author_css, el)
        publisher = self._css(self.metadata_publisher_css, el)
        journal = self._css(self.metadata_journal_css, el)
        date = self._css(self.metadata_date_css, el)
        language = self._css(self.metadata_language_css, el)
        volume = self._css(self.metadata_volume_css, el)
        issue = self._css(self.metadata_issue_css, el)
        firstpage = self._css(self.metadata_firstpage_css, el)
        lastpage = self._css(self.metadata_lastpage_css, el)
        doi = self._css(self.metadata_doi_css, el)
        pdf_url = self._css(self.metadata_pdf_url_css, el)
        html_url = self._css(self.metadata_html_url_css, el)

        metadata = {
                '_title': title[0].text if title else None,
                '_authors': [i.text for i in authors] if authors else None,
                '_publisher': publisher[0].text if publisher else None,
                '_journal': journal[0].text if journal else None,
                '_date': date[0].text if date else None,
                '_language': language[0].text if language else None,
                '_volume': volume[0].text if volume else None,
                '_issue': issue[0].text if issue else None,
                '_firstpage': firstpage[0].text if firstpage else None,
                '_lastpage': lastpage[0].text if lastpage else None,
                '_doi': doi[0].text if doi else None,
                '_pdf_url': self.url_prefix + pdf_url[0].text if pdf_url else None,
                '_html_url': self.url_prefix + html_url[0].text if html_url else None
        }
        meta = MetaData(metadata)
        return [meta]
